class ValidationError(Exception):
    """Error raised in validating Invoke context variables."""

    pass


class UndefinedVariable(ValidationError):
    """Error raised if a context variable remains undefined."""

    pass


class PathNotFound(ValidationError):
    """Error raised if a filesytem path isn't valid."""

    pass


class InvalidConfig(ValidationError):
    """Error raised in validating a config file schema."""

    pass



from invoke import Context, task

from noos_inv import exceptions


CONFIG = {
    "terraform": {
        "organisation": None,
        "workspace": None,
        "token": None,
    }
}


# Terraform deployment workflow


@task()
def update(
    ctx: Context,
    variable: str = "",
    value: str = "",
    organisation: str | None = None,
    workspace: str | None = None,
    token: str | None = None,
) -> None:
    """Update variable in Terraform cloud."""
    cmd = f"noostf update --variable {variable} --value '{value}'"
    ctx.run(cmd + _append_credentials(ctx, organisation, workspace, token), pty=True)


@task()
def run(
    ctx,
    message: str = "",
    organisation: str | None = None,
    workspace: str | None = None,
    token: str | None = None,
) -> None:
    """Run a plan in Terraform cloud."""
    cmd = f"noostf run --message '{message}'"
    ctx.run(cmd + _append_credentials(ctx, organisation, workspace, token), pty=True)


def _append_credentials(
    ctx: Context, organisation: str | None, workspace: str | None, token: str | None
) -> str:
    cmd = ""
    for arg in ctx.terraform:
        # Check credentials
        secret = locals()[arg] or ctx.terraform[arg]
        if secret is None:
            raise exceptions.UndefinedVariable(f"Missing Terraform Cloud variable {arg}")
        # Return credentials args
        cmd += f" --{arg} {secret}"
    return cmd



# Using `invoke` as a library
# http://docs.pyinvoke.org/en/stable/concepts/library.html


from invoke import Collection, Config, Program

from noos_inv import __version__
from noos_inv.tasks import docker, git, helm, local, python, terraform


class BaseConfig(Config):
    prefix = "noosinv"


namespace = Collection()
namespace.add_collection(Collection.from_module(docker, config=docker.CONFIG))
namespace.add_collection(Collection.from_module(git, config=git.CONFIG))
namespace.add_collection(Collection.from_module(helm, config=helm.CONFIG))
namespace.add_collection(Collection.from_module(local, config=local.CONFIG))
namespace.add_collection(Collection.from_module(python, config=python.CONFIG))
namespace.add_collection(Collection.from_module(terraform, config=terraform.CONFIG))


program = Program(namespace=namespace, config_class=BaseConfig, version=__version__)



import pathlib
from collections.abc import Mapping

from noos_inv import exceptions, types


def check_path(path: str) -> None:
    """Check whether a path exists on a file system."""
    if not pathlib.Path(path).exists():
        raise exceptions.PathNotFound(f"Incorrect file system path: {path}")


def check_config(config: object) -> None:
    """Check whether a port-forward configuration is valid."""
    if not isinstance(config, Mapping):
        raise exceptions.InvalidConfig("Configuration must be an mapping")
    if len(config) == 0:
        raise exceptions.InvalidConfig("Configuration can not be empty")
    for item in config.values():
        if not isinstance(item, Mapping):
            raise exceptions.InvalidConfig("Configuration element must be a mapping")
        if not (types.PodConfig.__required_keys__ <= set(item.keys())):
            raise exceptions.InvalidConfig("Invalid configuration element keys")



from enum import StrEnum, auto
from typing import NotRequired, TypedDict

from noos_inv import exceptions


class PodConfig(TypedDict):
    podNamespace: str
    podPrefix: str
    podPort: int
    localPort: int
    localAddress: NotRequired[str]


type PodsConfig = dict[str, PodConfig]


class ValidatedEnum(StrEnum):
    """Specific Enum with a validated getter method."""

    @classmethod
    def get(cls, value: str) -> StrEnum:
        if value not in cls:
            raise exceptions.UndefinedVariable(f"Unknown {cls.__name__} {value}")
        return cls(value)


class UserType(StrEnum):
    AWS = "AWS"


class InstallType(ValidatedEnum):
    PIPENV = auto()
    POETRY = auto()
    UV = auto()


class GroupType(ValidatedEnum):
    UNIT = auto()
    INTEGRATION = auto()
    FUNCTIONAL = auto()


class FormatterType(ValidatedEnum):
    BLACK = auto()
    ISORT = auto()
    RUFF = auto()


class LinterType(ValidatedEnum):
    BLACK = auto()
    ISORT = auto()
    PYDOCSTYLE = auto()
    FLAKE8 = auto()
    RUFF = auto()
    MYPY = auto()
    IMPORTS = auto()



import json
import logging
import pathlib

from invoke import Context, task

from noos_inv import exceptions, types, validators


logger = logging.getLogger(__name__)


CONFIG = {
    "local": {
        # Sensitive
        "config": None
    }
}


# Local development workflow


@task(help={"force": "Whether to destroy the existing file first"})
def dotenv(
    ctx: Context,
    template: str = "./dotenv.tpl",
    target: str = "./.env",
    force: bool = False,
) -> None:
    """Create local dotenv file."""
    validators.check_path(template)
    try:
        validators.check_path(target)
        if force:
            raise exceptions.PathNotFound
    except exceptions.PathNotFound:
        ctx.run(f"cp {template} {target}")


@task(
    help={
        "pod": "Forward port only for this specific pod",
        "unforward": "Unforward ports without forwarding them again",
        "config": "Configuration file path (including pods to port-forward)",
    }
)
def ports(
    ctx: Context,
    pod: str | None = None,
    unforward: bool = False,
    config: str | None = None,
) -> None:
    """Forward ports for defined Kubernetes pods."""
    config = config or ctx.local.config
    if config is None:
        raise exceptions.UndefinedVariable("Missing local config file")
    # Load config file
    validators.check_path(config)
    with pathlib.Path(config).open(mode="rt") as f:
        local_config: types.PodsConfig = json.load(f).get("podForwards")
    validators.check_config(local_config)
    # Narrow-down config if necessary
    tmp_config: types.PodsConfig
    if pod is None:
        tmp_config = local_config
    else:
        if pod not in local_config:
            raise exceptions.UndefinedVariable("Missing pod in config file")
        tmp_config = {pod: local_config[pod]}
    # Iterate over targeted services
    filtered_pods = _filter_pods(ctx, tmp_config)
    for pod, pod_config in tmp_config.items():
        if unforward:
            # Unforward port
            _unforward(ctx, pod_config)
        else:
            # Forward port
            _forward(ctx, pod_config, filtered_pods[pod])


def _filter_pods(ctx: Context, config: types.PodsConfig) -> dict[str, str]:
    """Filter all matching pods in a given namespace."""
    # Query all pods in the namespace
    cmd_tpl = "kubectl get pod -n {namespace} "
    # Select only the name
    cmd_tpl += "-o=custom-columns=NAME:.metadata.name "
    # Filter out pods with the prefix
    cmd_tpl += " | grep {prefix}"
    # Build data struct {service: pod_name}
    selected_pods: dict[str, str] = {}
    for pod, pod_config in config.items():
        cmd = cmd_tpl.format(
            namespace=pod_config["podNamespace"],
            prefix=pod_config["podPrefix"],
        )
        result = ctx.run(cmd, hide=True)
        if result is None:
            logger.error(f"Failed to fetch pod name for {pod}. Skip!")
            continue
        selected_pods[pod] = result.stdout.rstrip()
    # Return selected pod names for each service
    return selected_pods


def _get_kubectl_command(
    *, namespace: str, name: str, port: int, local_port: int, local_address: str | None = None
) -> str:
    """Get the command to forward a port to a pod."""
    cmd = f"kubectl port-forward -n {namespace} {name} {local_port}:{port}"
    if local_address is not None:
        cmd += f" --address={local_address}"
    return cmd


def _forward(ctx: Context, config: types.PodConfig, pod_name: str) -> None:
    """Forward port matching configuration."""
    # Build kubectl port-forward command
    cmd = _get_kubectl_command(
        namespace=config["podNamespace"],
        name=pod_name,
        port=config["podPort"],
        local_port=config["localPort"],
        local_address=config.get("localAddress"),
    )
    # Ensure the process is detached
    cmd += " </dev/null >/dev/null 2>&1 &"
    logger.warning(f"Forwarding {config['podNamespace']}/{pod_name} to :{config['localPort']}")
    ctx.run(cmd, warn=True, hide=True)


def _unforward(ctx: Context, config: types.PodConfig) -> None:
    """Unforward port matching configuration."""
    # Build kubectl port-forward command
    cmd = _get_kubectl_command(
        namespace=config["podNamespace"],
        name=config["podPrefix"] + ".*",
        port=config["podPort"],
        local_port=config["localPort"],
    )
    # Fetch running processes
    cmd = f"ps aux | grep '{cmd}' | grep -v grep"
    # Restrict to only pid part
    cmd += " | awk '{print $2}'"
    result = ctx.run(cmd, warn=True, hide=True)
    # Kill the process
    if result is not None:
        if result.stdout != "":
            logger.warning(f"Killing port-forward at :{config['localPort']}")
            ctx.run(f"kill -9 {result.stdout.rstrip()}", warn=True, hide=True)



from invoke import Context, task

from noos_inv import exceptions, types, validators


CONFIG = {
    "python": {
        "install": "uv",
        "source": "./src",
        "formatters": "ruff",
        "linters": "ruff,mypy",
        "tests": "./src/tests",
        "user": None,
        "token": None,
    }
}


# Python deployment workflow


@task()
def clean(ctx: Context) -> None:
    """Clean project from temp files / dirs."""
    ctx.run("rm -rf build dist")
    ctx.run("find src -type d -name __pycache__ | xargs rm -rf")


@task()
def format(
    ctx: Context,
    formatters: str | None = None,
    source: str | None = None,
    install: str | None = None,
) -> None:
    """Auto-format source code."""
    formatters = formatters or ctx.python.formatters
    source = source or ctx.python.source
    validators.check_path(source)
    cmd = _activate_shell(ctx, install)
    for formatter in formatters.split(","):
        match types.FormatterType.get(formatter):
            case types.FormatterType.BLACK:
                ctx.run(cmd + f"black {source}", pty=True)
            case types.FormatterType.ISORT:
                ctx.run(cmd + f"isort {source}", pty=True)
            case types.FormatterType.RUFF:
                ctx.run(cmd + f"ruff check --select I --fix {source}")
                ctx.run(cmd + f"ruff format {source}")


@task()
def lint(
    ctx: Context,
    linters: str | None = None,
    source: str | None = None,
    install: str | None = None,
) -> None:
    """Run python linters."""
    linters = linters or ctx.python.linters
    source = source or ctx.python.source
    validators.check_path(source)
    cmd = _activate_shell(ctx, install)
    for linter in linters.split(","):
        match types.LinterType.get(linter):
            case types.LinterType.BLACK:
                ctx.run(cmd + f"black --check {source}", pty=True)
            case types.LinterType.ISORT:
                ctx.run(cmd + f"isort --check-only {source}", pty=True)
            case types.LinterType.PYDOCSTYLE:
                ctx.run(cmd + f"pydocstyle {source}", pty=True)
            case types.LinterType.FLAKE8:
                ctx.run(cmd + f"flake8 {source}", pty=True)
            case types.LinterType.MYPY:
                ctx.run(cmd + f"mypy {source}", pty=True)
            case types.LinterType.RUFF:
                ctx.run(cmd + f"ruff check {source}")
            case types.LinterType.IMPORTS:
                ctx.run(cmd + "lint-imports", pty=True)


@task()
def test(
    ctx: Context, tests: str | None = None, group: str = "", install: str | None = None
) -> None:
    """Run pytest with optional grouped tests."""
    tests = tests or ctx.python.tests
    if group != "":
        tests += "/" + types.GroupType.get(group)
    validators.check_path(tests)
    cmd = _activate_shell(ctx, install)
    ctx.run(cmd + f"pytest {tests}", pty=True)


@task()
def coverage(
    ctx: Context,
    config: str = "setup.cfg",
    report: str = "term",
    tests: str | None = None,
    install: str | None = None,
) -> None:
    """Run coverage test report."""
    tests = tests or ctx.python.tests
    validators.check_path(tests)
    cmd = _activate_shell(ctx, install)
    ctx.run(cmd + f"pytest --cov --cov-config={config} --cov-report={report} {tests}", pty=True)


@task()
def package(ctx: Context, install: str | None = None) -> None:
    """Build project wheel distribution."""
    install = install or ctx.python.install
    match types.InstallType.get(install):
        case types.InstallType.POETRY:
            ctx.run("poetry build", pty=True)
        case types.InstallType.PIPENV:
            ctx.run("pipenv run python -m build -n", pty=True)
        case types.InstallType.UV:
            ctx.run("uvx --from build pyproject-build --installer uv")


@task()
def release(
    ctx: Context, user: str | None = None, token: str | None = None, install: str | None = None
) -> None:
    """Publish wheel distribution to PyPi."""
    user = user or ctx.python.user
    token = token or ctx.python.token
    install = install or ctx.python.install
    if user is None:
        raise exceptions.UndefinedVariable("Missing remote PyPi registry user")
    if token is None:
        raise exceptions.UndefinedVariable("Missing remote PyPi registry token")
    match types.InstallType.get(install):
        case types.InstallType.POETRY:
            ctx.run(f"poetry publish --build -u {user} -p {token}", pty=True)
        case types.InstallType.PIPENV:
            ctx.run(f"pipenv run twine upload dist/* -u {user} -p {token}", pty=True)
        case types.InstallType.UV:
            ctx.run(f"uvx twine upload dist/* -u {user} -p {token}")


def _activate_shell(ctx: Context, install: str | None) -> str:
    install = install or ctx.python.install
    return f"{types.InstallType.get(install)} run "
